% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file err.pl
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% error handling 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
compilation_error(parsing(_, _, _)).
compilation_error(generic_rewriting(_, _, _)).
compilation_error(code_generation(_, _, _)).

error_types([parsing, generic_rewriting, code_generation]).

error_message(not_parsed_as, "\"~w\" not parsed as ~w").
error_message(not_defined, "\"~w\" is unknown\n").
error_message(not_reduced_to_type, "\"~w\" should had been reduced to \"~w\"\n").
error_message(not_allowed_in, "\"~w\" not allowed in \"~w\" mode\n").
error_message(not_reduced_to_term, "\"~w\" should had been reduced to term \"~w\"\n").
error_message(not_accepted_by_language, "\"~w\" is not accepted by \"~w\" language\n").
error_message(not_bounded_variable, "unbound variable \"~w\" (consider using the dynamic predicate)\n").

error(Type, String, Terms) :-
	write('---! \"'),
	write(Type),
	write(' error'),
	write('\" !--- \n\t'),
	format(String, Terms),
	nl,nl,
	raise_exception(error).

error_handler(Error) :-
	print_message(error, Error).
  
portray_message(error, MessageTerm) :-
	MessageTerm =.. [ErrType, SubMessageTerm, InvolvedTerms, CallStack],
	error_types(ErrTypes),
	memberchk(ErrType, ErrTypes),
	SubMessageTerm =.. [ErrSubType | InvolvedSubTerms],
	error_message(ErrSubType, String),
	append(InvolvedTerms, InvolvedSubTerms, Terms),	
	format("! ~w error !\n! ", [ErrType]),	
	format(String, Terms),nl,
	write('----------'),nl,
	write('call_stack :'),nl,
	write('----------'),nl,
	write_list(CallStack),nl,
	halt.

syntax_error(Stream, Char, Explain) :-
   atom_concat([Char, Explain], Msg),
   line_count(Stream, LineMinusOne),
   Line is LineMinusOne + 1,
   throw(error(syntax_error(Msg), syntax_error(parse_stream(_), Line, Msg, [Char - Line], 1))).