:- use_module(library(clpfd)).
:- use_module('/afs/inria.fr/rocq/home/contraintes/common/src/rules2cp/interpreter/prolog/r2cp.pl').
:- init_interpreter.


rcp_q(I, _A_) :- 
	rcp_var(from(q(I),0,1), Q_I),
	rcp_rec([row=Q_I,column=I], _A_).

rcp_q_search(I, _A_) :- 
	rcp_var(from(q(I),0,1), Q_I),
	rcp_rec([row=Q_I,column=I], _A_).

rcp_board_foldr_1([], _, []) :- 
	!.
rcp_board_foldr_1([I_69|Tail], [N], [_A_|_B_]) :- 
	rcp_q(I_69, _A_),
	rcp_board_foldr_1(Tail, [N], _B_).

rcp_board_foldr_1_search([], _, []) :- 
	!.
rcp_board_foldr_1_search([I_69|Tail], [N], [_C_|_D_]) :- 
	rcp_q_search(I_69, _C_),
	rcp_board_foldr_1_search(Tail, [N], _D_).

rcp_board(N, _B_) :- 
	rcp_range([1..N], _A_),
	rcp_board_foldr_1(_A_, [N], _B_).

rcp_board_search(N, _B_) :- 
	rcp_range([1..N], _A_),
	rcp_board_foldr_1_search(_A_, [N], _B_).

rcp_safe_foldr_3([], _, 1) :- 
	!.
rcp_safe_foldr_3([R_74|Tail], [L,Q_70], _Q_) :- 
	fd_and(_O_, _P_, _Q_),
	rcp_att(column, Q_70, I_77),
	rcp_att(column, R_74, J_79),
	fd_implies(_C_, _N_, _O_),
	fd_lt(I_77, J_79, _C_),
	fd_and(_H_, _M_, _N_),
	rcp_att(row, Q_70, _D_),
	fd_aeq(J_79-I_77, _E_),
	rcp_att(row, R_74, _F_),
	fd_aeq(_E_+_F_, _G_),
	fd_neq(_D_, _G_, _H_),
	rcp_att(row, Q_70, _I_),
	fd_aeq(I_77-J_79, _J_),
	rcp_att(row, R_74, _K_),
	fd_aeq(_J_+_K_, _L_),
	fd_neq(_I_, _L_, _M_),
	rcp_safe_foldr_3(Tail, [L,Q_70], _P_).

rcp_safe_foldr_3_search([], _, 1) :- 
	!.
rcp_safe_foldr_3_search([R_74|Tail], [L,Q_70], _) :- 
	rcp_att(column, Q_70, I_77),
	rcp_att(column, R_74, J_79),
	(
	fd_lt(I_77, J_79),
	rcp_att(row, Q_70, _T_),
	fd_aeq(J_79-I_77, _U_),
	rcp_att(row, R_74, _V_),
	fd_aeq(_U_+_V_, _W_),
	rcp_att(row, Q_70, _X_),
	fd_aeq(I_77-J_79, _Y_),
	rcp_att(row, R_74, _Z_),
	fd_aeq(_Y_+_Z_, _B1_),
	fd_equiv((fd_neq(_T_,_W_),fd_neq(_X_,_B1_)), _C1_, 1)
	;
	fd_geq(I_77, J_79),
	fd_aeq(_C1_, 1)
	),
	rcp_assert(_C1_),
	rcp_safe_foldr_3_search(Tail, [L,Q_70], _D1_),
	rcp_assert(_D1_).

rcp_safe_foldr_2([], _, 1) :- 
	!.
rcp_safe_foldr_2([Q_70|Tail], [L], _C_) :- 
	fd_and(_A_, _B_, _C_),
	rcp_safe_foldr_3(L, [L,Q_70], _A_),
	rcp_safe_foldr_2(Tail, [L], _B_).

rcp_safe_foldr_2_search([], _, 1) :- 
	!.
rcp_safe_foldr_2_search([Q_70|Tail], [L], _) :- 
	rcp_safe_foldr_3_search(L, [L,Q_70], _D_),
	rcp_assert(_D_),
	rcp_safe_foldr_2_search(Tail, [L], _E_),
	rcp_assert(_E_).

rcp_safe(L, _B_) :- 
	variables(L, _A_),
	all_different(_A_),
	rcp_safe_foldr_2(L, [L], _B_).

rcp_safe_search(L, _C_) :- 
	variables(L, _A_),
	all_different(_A_),
	rcp_safe_foldr_2_search(L, [L], _B_),
	rcp_assert(_B_),
	fd_aeq(_C_, 1).

rcp_queens_constraints(B, N, _B_) :- 
	variables(B, _A_),
	domain(_A_, 1, N),
	rcp_safe(B, _B_).

rcp_queens_constraints_search(B, N, _C_) :- 
	variables(B, _A_),
	domain(_A_, 1, N),
	rcp_safe_search(B, _B_),
	rcp_assert(_B_),
	fd_aeq(_C_, 1).

rcp_var_order_criterion(1, Candidate, [], _C_) :- 
	rcp_call(Candidate, _A_),
	rcp_att(row, _A_, _B_),
	fd_size(_B_, _C_).

rcp_queens_search(B, _B_) :- 
	variables(B, _A_),
	variable_ordering([least(rcp_var_order_criterion(1,[]))]),
	rcp_labeling(_A_),
	fd_aeq(_B_, 1).

rcp_queens_search_search(B, _B_) :- 
	variables(B, _A_),
	variable_ordering([least(rcp_var_order_criterion(1,[]))]),
	rcp_labeling(_A_),
	fd_aeq(_B_, 1).

solve([Q_1,Q_2,Q_3,Q_4]) :- 
	statistics(runtime, [_Runtime_0_|_]),
	fd_statistics(backtracks, _),
	rcp_var(from(q(1),0,1), Q_1),
	rcp_var(from(q(2),0,1), Q_2),
	rcp_var(from(q(3),0,1), Q_3),
	rcp_var(from(q(4),0,1), Q_4),
	domain([Q_1,Q_2,Q_3,Q_4], 1, 4),
	all_different([Q_1,Q_2,Q_3,Q_4]),
	Q_1#\=1+Q_2,
	Q_1#\= -1+Q_2,
	Q_1#\=2+Q_3,
	Q_1#\= -2+Q_3,
	Q_1#\=3+Q_4,
	Q_1#\= -3+Q_4,
	Q_2#\=1+Q_3,
	Q_2#\= -1+Q_3,
	Q_2#\=2+Q_4,
	Q_2#\= -2+Q_4,
	Q_3#\=1+Q_4,
	Q_3#\= -1+Q_4,
	rcp_var(from(q(1),0,1), Q_1),
	rcp_rec([row=Q_1,column=1,uid=1], _A_),
	rcp_var(from(q(2),0,1), Q_2),
	rcp_rec([row=Q_2,column=2,uid=2], _B_),
	rcp_var(from(q(3),0,1), Q_3),
	rcp_rec([row=Q_3,column=3,uid=3], _C_),
	rcp_var(from(q(4),0,1), Q_4),
	rcp_rec([row=Q_4,column=4,uid=4], _D_),
	rcp_queens_search([_A_,_B_,_C_,_D_], _E_),
	rcp_assert(_E_),
	statistics(runtime, [_Runtime_1_|_]),
	fd_statistics(backtracks, _Backtracks_),
	_Runtime_ is _Runtime_1_-_Runtime_0_,
	format('~3d seconds\n', [_Runtime_]),
	format('~w bactracks\n\n', [_Backtracks_]).
