:- use_module(library(clpfd)).
:- use_module('/afs/inria.fr/rocq/home/contraintes/common/src/rules2cp/interpreter/prolog/r2cp.pl').
:- init_interpreter.



rcp_t1(_A_) :- 
	rcp_var(from(t1,0,1), T1),
	rcp_rec([start=T1,duration=2], _A_).

rcp_t1_search(_A_) :- 
	rcp_var(from(t1,0,1), T1),
	rcp_rec([start=T1,duration=2], _A_).

rcp_t2(_A_) :- 
	rcp_var(from(t2,0,1), T2),
	rcp_rec([start=T2,duration=5], _A_).

rcp_t2_search(_A_) :- 
	rcp_var(from(t2,0,1), T2),
	rcp_rec([start=T2,duration=5], _A_).

rcp_t3(_A_) :- 
	rcp_var(from(t3,0,1), T3),
	rcp_rec([start=T3,duration=4], _A_).

rcp_t3_search(_A_) :- 
	rcp_var(from(t3,0,1), T3),
	rcp_rec([start=T3,duration=4], _A_).

rcp_t4(_A_) :- 
	rcp_var(from(t4,0,1), T4),
	rcp_rec([start=T4,duration=3], _A_).

rcp_t4_search(_A_) :- 
	rcp_var(from(t4,0,1), T4),
	rcp_rec([start=T4,duration=3], _A_).

rcp_t5(_A_) :- 
	rcp_var(from(t5,0,1), T5),
	rcp_rec([start=T5,duration=1], _A_).

rcp_t5_search(_A_) :- 
	rcp_var(from(t5,0,1), T5),
	rcp_rec([start=T5,duration=1], _A_).

rcp_cost(_B_) :- 
	t5(_A_),
	rcp_att(start, _A_, _B_).

rcp_cost_search(_B_) :- 
	rcp_t5_search(_A_),
	rcp_att(start, _A_, _B_).

rcp_precedences(_W_) :- 
	fd_and(_C_, _V_, _W_),
	t1(_A_),
	t2(_B_),
	rcp_prec(_A_, _B_, _C_),
	fd_and(_F_, _U_, _V_),
	t2(_D_),
	t5(_E_),
	rcp_prec(_D_, _E_, _F_),
	fd_and(_I_, _T_, _U_),
	t1(_G_),
	t3(_H_),
	rcp_prec(_G_, _H_, _I_),
	fd_and(_L_, _S_, _T_),
	t3(_J_),
	t5(_K_),
	rcp_prec(_J_, _K_, _L_),
	fd_and(_O_, _R_, _S_),
	t1(_M_),
	t3(_N_),
	rcp_prec(_M_, _N_, _O_),
	t3(_P_),
	t5(_Q_),
	rcp_prec(_P_, _Q_, _R_).

rcp_precedences_search(_S_) :- 
	rcp_t1_search(_A_),
	rcp_t2_search(_B_),
	rcp_prec_search(_A_, _B_, _C_),
	rcp_assert(_C_),
	rcp_t2_search(_D_),
	rcp_t5_search(_E_),
	rcp_prec_search(_D_, _E_, _F_),
	rcp_assert(_F_),
	rcp_t1_search(_G_),
	rcp_t3_search(_H_),
	rcp_prec_search(_G_, _H_, _I_),
	rcp_assert(_I_),
	rcp_t3_search(_J_),
	rcp_t5_search(_K_),
	rcp_prec_search(_J_, _K_, _L_),
	rcp_assert(_L_),
	rcp_t1_search(_M_),
	rcp_t3_search(_N_),
	rcp_prec_search(_M_, _N_, _O_),
	rcp_assert(_O_),
	rcp_t3_search(_P_),
	rcp_t5_search(_Q_),
	rcp_prec_search(_P_, _Q_, _R_),
	rcp_assert(_R_),
	fd_aeq(_S_, 1).

rcp_disjunctives(_K_) :- 
	fd_and(_C_, _J_, _K_),
	t2(_A_),
	t4(_B_),
	rcp_disj(_A_, _B_, _C_),
	fd_and(_F_, _I_, _J_),
	t3(_D_),
	t4(_E_),
	rcp_disj(_D_, _E_, _F_),
	t2(_G_),
	t3(_H_),
	rcp_disj(_G_, _H_, _I_).

rcp_disjunctives_search(_J_) :- 
	rcp_t2_search(_A_),
	rcp_t4_search(_B_),
	rcp_disj_search(_A_, _B_, _C_),
	rcp_assert(_C_),
	rcp_t3_search(_D_),
	rcp_t4_search(_E_),
	rcp_disj_search(_D_, _E_, _F_),
	rcp_assert(_F_),
	rcp_t2_search(_G_),
	rcp_t3_search(_H_),
	rcp_disj_search(_G_, _H_, _I_),
	rcp_assert(_I_),
	fd_aeq(_J_, 1).

rcp_prec(T1, T2, _E_) :- 
	rcp_att(start, T1, _A_),
	rcp_att(duration, T1, _B_),
	fd_aeq(_A_+_B_, _C_),
	rcp_att(start, T2, _D_),
	fd_leq(_C_, _D_, _E_).

rcp_prec_search(T1, T2, _E_) :- 
	rcp_att(start, T1, _A_),
	rcp_att(duration, T1, _B_),
	fd_aeq(_A_+_B_, _C_),
	rcp_att(start, T2, _D_),
	fd_leq(_C_, _D_),
	fd_aeq(_E_, 1).

rcp_disj(T1, T2, _C_) :- 
	fd_or(_A_, _B_, _C_),
	rcp_prec(T1, T2, _A_),
	rcp_prec(T2, T1, _B_).

rcp_disj_search(T1, T2, _C_) :- 
	(
	rcp_prec_search(T1, T2, _A_),
	rcp_assert(_A_)
	;
	rcp_prec_search(T2, T1, _B_),
	rcp_assert(_B_)
	),
	fd_aeq(_C_, 1).

solve([T1,T2,T4,T3,T5]) :- 
	statistics(runtime, [_Runtime_0_|_]),
	fd_statistics(backtracks, _),
	rcp_var(from(t1,0,1), T1),
	rcp_var(from(t5,0,1), T5),
	rcp_var(from(t2,0,1), T2),
	rcp_var(from(t4,0,1), T4),
	rcp_var(from(t3,0,1), T3),
	domain([T1,T2,T3,T4,T5], 0, 20),
	T1+2#=<T2,
	T2+5#=<T5,
	T1+2#=<T3,
	T3+4#=<T5,
	T1+2#=<T3,
	T3+4#=<T5,
	minimize((((T2+5#=<T3;T3+4#=<T2),(T2+5#=<T4;T4+3#=<T2),(T3+4#=<T4;T4+3#=<T3)),labeling([up],[T5])), T5),
	statistics(runtime, [_Runtime_1_|_]),
	fd_statistics(backtracks, _Backtracks_),
	_Runtime_ is _Runtime_1_-_Runtime_0_,
	format('~3d seconds\n', [_Runtime_]),
	format('~w bactracks\n\n', [_Backtracks_]).
