:- use_module(library(codesio)).

decompose_big([], _X, _T, _Op, _Labels, []).
decompose_big([H | L], X, T, Op, Labels, Children) :-
	copy_term((T, X), (T0, H)),
	decompose(T0, Op, Labels, Children0),
	decompose_big(L, X, T, Op, Labels, Children1),
	append(Children0, Children1, Children).


decompose_term(big(X, L, T), Op, Labels, Children) :-
	decompose_big(L, X, T, Op, Labels, Children).
decompose_term(pair(A, B), Op, Labels, Children) :-
	decompose(A, Op, Labels, ChildrenA),
	decompose(B, Op, Labels, ChildrenB),
	append(ChildrenA, ChildrenB, Children).

decompose(op(Op, Term), Op, Labels, Children) :-
	!,
	decompose_term(Term, Op, Labels, Children).
decompose(label(Label, Term), Op, Labels, Children) :-
	decompose(Term, Op, [Label | Labels], Children).
decompose(Term, _, Labels, [Labels - Term]).

connect_op(and, A, B, (A, B)).
connect_op(or, A, B, (A; B)).
	

rewrite(Term, Criter, Counter0, Clauses, Root, Counter1) :-
	decompose(Term, Op, [], Children),
	(
	  Op = leaf
	->
	  Children = [Root],
	  Clauses = [],
	  Counter1 = Counter0
	;
	  Ncounter is Counter0 + 1,
	  format_to_codes('p~d', [Counter0], Functor_codes),
	  atom_codes(Functor, Functor_codes),
	  rewrite_list(Children, Criter, Ncounter, Clauses0, Roots, Counter1),
	  Init =.. [Functor, Roots],
	
	  Root = Functor,
	  Aux =.. [Functor, List],
	
	  Recurse =.. [Functor, T],
	  connect_op(Op, call(H), Recurse, Body),
	
	  Clauses = [(Aux :- leader(List, Criter, H, T), Body),
		     (Root :- Init)
		    | Clauses0]
	).

rewrite_list([], _, Counter, [], [], Counter).
rewrite_list([Labels - H | T], Criterion, Counter0, Clauses, Roots, Counter1) :-
	rewrite(H, Criterion, Counter0, Clauses0, Root, Counter),
	Roots = [Labels - Root | Tail],
	append(Clauses0, Clauses1, Clauses),
	rewrite_list(T, Criterion, Counter, Clauses1, Tail, Counter1).

test(A, B, C, Clauses, Root) :-
	rewrite(op(and, big(X, [A, B, C], label(var(X), op(or, big(V, [1, 2, 3], X = V))))),
		[],
		0,
		Clauses,
		Root,
		_).


