r2cp_builtin_type('true', [], value(bool)).
r2cp_builtin_type('false', [], value(bool)).
r2cp_builtin_type('+', [value(int), value(int)], value(int)).
r2cp_builtin_type('-', [value(int), value(int)], value(int)).
r2cp_builtin_type('*', [value(int), value(int)], value(int)).
r2cp_builtin_type('/', [value(int), value(int)], value(int)).
r2cp_builtin_type('min', [value(int), value(int)], value(int)).
r2cp_builtin_type('max', [value(int), value(int)], value(int)).
r2cp_builtin_type('log', [value(int), value(int)], value(int)).
r2cp_builtin_type('exp', [value(int), value(int)], value(int)).
r2cp_builtin_type('=', [value(int), value(int)], value(bool)).
r2cp_builtin_type('<', [value(int), value(int)], value(bool)).
r2cp_builtin_type('>', [value(int), value(int)], value(bool)).
r2cp_builtin_type('=<', [value(int), value(int)], value(bool)).
r2cp_builtin_type('>=', [value(int), value(int)], value(bool)).
r2cp_builtin_type('#', [value(int), value(int)], value(bool)).
r2cp_builtin_type('and', [value(bool), value(bool)], value(bool)).
r2cp_builtin_type('or', [value(bool), value(bool)], value(bool)).
r2cp_builtin_type('not', [value(bool)], value(bool)).
r2cp_builtin_type('implies', [value(bool), value(bool)], value(bool)).
r2cp_builtin_type('equiv', [value(bool), value(bool)], value(bool)).
r2cp_builtin_type('nth', [value(int), [A]], A).

% r2cp_definition(end, [var('Task', 0)],
%    (var('Task', 0):start) + (var('Task', 0):duration), _).
% r2cp_definition(prec, [var('A', 0), var('B', 1)],
%    end(var('A', 0)) < (var('B', 1):start), _).
% r2cp_definition(disj, [var('A', 0), var('B', 1)],
%    or(
%       prec(var('A', 0), var('B', 1)),
%       prec(var('B', 0), var('A', 1))), _).
% r2cp_definition(disj_list, [var('Tasks', 0)],
%    forall(var('X', 0), var('Tasks', 0),
%       forall(var('Y', 0), var('Tasks', 0),
%          implies(
%             #(var('X', 0):id, var('Y', 0):id),
%             disj(var('X', 0), var('Y', 0))))), _).

% r2cp_goal(
%    disj_list([
%       { id = 0, start = 1, duration = 2},
%       { id = 1, start = 6, duration = 3}]), _).

r2cp_definition(Func, Args, Body) :-
   rule(Head, Body),
   !,
   Head =.. [Func | Args].
r2cp_definition(Func, Args, Body) :-
   function(Head, Body),
   !,
   Head =.. [Func | Args].
r2cp_definition(Func, [], Body) :-
   constant(Func, Body).

r2cp_goal(G) :- query(G).

check_types :-
   retractall(r2cp_var_counter(_)),
   retractall(r2cp_known_type(_, _, _)),
   retractall(r2cp_type_being_computed(_)),
   (
      r2cp_definition(F, _, _, _),
      (
         get_type(F, _, _, TyArgs, Ty),
         concrete_type((TyArgs, Ty)),
         P =.. [F | TyArgs],
         print(P : Ty),
         nl
      ->
         fail
      )
   ;
      true
   ),
   r2cp_goal(G, Pos),
   type_expr(G, _, Pos, Ty),
   unify_type(Ty, value(bool), _, G).

get_type(F, Pos, T, TyArgs, Ty) :-
   (
      r2cp_builtin_type(F, TyArgs0, Ty0)
   ->
      true
   ;  
      r2cp_known_type(F, TyArgs0, Ty0)
   ->
      true
   ;
      r2cp_type_being_computed(F)
   ->
      throw(type_error('Recursive call'(F), Pos, T))
   ;
      r2cp_definition(F, Args, Body, Pos0)
   ->
      assert(r2cp_type_being_computed(F)),
      type_list_aux(Args, Ctx, Pos0, TyArgs0),
      type_expr(Body, Ctx, Pos0, Ty0),
      concrete_type((TyArgs0, Ty0)),
      assert(r2cp_known_type(F, TyArgs0, Ty0)),
      retract(r2cp_type_being_computed(F))
   ;
      throw(type_error('Unknown call'(F), Pos, T))
   ),
   generalize_type(Ty0, Ty1),
   copy_term((TyArgs0, Ty1), (TyArgs, Ty)).

concrete_type(Ty) :-
   (
      var(Ty)
   ->
      Ty = Ty
   ;
      (
         Ty = value(bool)
      ;
         Ty = value(int)
      )
   ->
      true
   ;
      callable(Ty)
   ->
      Ty =.. [_F | Args],
      concrete_type_args(Args)
   ).

concrete_type_args([]).
concrete_type_args([H | T]) :-
   concrete_type(H),
   concrete_type_args(T).

generalize_type(Ty0, Ty1) :-
   (
      var(Ty0)
   ->
      Ty1 = Ty0
   ;
      Ty0 = value(bool)
   ->
      Ty1 = value(_)
   ;
      callable(Ty0)
   ->
      Ty0 =.. [F | Args0],
      generalize_type_args(Args0, Args1),
      Ty1 =.. [F | Args1]
   ;
      Ty1 = Ty0
   ).

generalize_type_args([], []).
generalize_type_args([H0 | T0], [H1 | T1]) :-
   generalize_type(H0, H1),
   generalize_type_args(T0, T1).

type_expr(N, _Ctx, _Pos, Ty) :-
   integer(N),
   !,
   Ty = value(int).
type_expr([], _Ctx, _Pos, Ty) :-
   !,
   Ty = [_].
type_expr([H | T], Ctx, Pos, Ty) :-
   !,
   type_expr(H, Ctx, Pos, TyH),
   type_expr(T, Ctx, Pos, TyT),
   unify_type(TyT, [TyH], Pos, [H | T]),
   Ty = TyT.
type_expr({}, _Ctx, _Pos, Ty) :-
   !,
   Ty = record([]).
type_expr({ R }, Ctx, Pos, Ty) :-
   !,
   type_record(R, Ctx, Pos, TyR),
   Ty = { TyR }.
type_expr(R:L, Ctx, Pos, Ty) :-
   !,
   type_expr(R, Ctx, Pos, TyR),
   unify_type(TyR, {[L = Ty | _]}, Pos, R:L).
type_expr(var(V, _), Ctx, _Pos, Ty) :-
   member(V - Ty, Ctx), !.
type_expr(Expr, Ctx, Pos, Ty) :-
   Expr = fold_left(OpEnv, List, X, I, E),
   !,
   OpEnv =.. [Op | Env],
   get_type(Op, Pos, OpEnv, TyArgs, Ty),
   new_var_name(Accu),
   new_var_name(V),
   append(Env, [var(Accu, _), var(V, _)], Args),
   type_list(Args, [Accu - Ty, V - TyV | Ctx], Op, Pos, Expr, TyArgs),
   type_expr(List, Ctx, Pos, TyList),
   unify_type(TyList, [ItemTy], Pos, List),
   (
      X = var(NameX, _)
   ->
      true
   ;
      throw(type_error('Should be a variable'(X)), Pos, Expr)
   ),
   type_expr(I, Ctx, Pos, TyI),
   unify_type(TyI, Ty, Pos, I),
   type_expr(E, [NameX - ItemTy | Ctx], Pos, TyE),
   unify_type(TyE, TyV, Pos, E).
type_expr(Expr, Ctx, Pos, Ty) :-
   Expr = fold_right(OpEnv, List, X, E, I),
   !,
   OpEnv =.. [Op | Env],
   get_type(Op, Pos, OpEnv, TyArgs, Ty),
   new_var_name(Accu),
   new_var_name(V),
   append(Env, [var(Accu, _), var(V, _)], Args),
   type_list(Args, [V - TyV, Accu - Ty| Ctx], Op, Pos, Expr, TyArgs),
   type_expr(List, Ctx, Pos, TyList),
   unify_type(TyList, [ItemTy], Pos, List),
   (
      X = var(NameX, _)
   ->
      true
   ;
      throw(type_error('Should be a variable'(X)), Pos, Expr)
   ),
   type_expr(E, [NameX - ItemTy | Ctx], Pos, TyE),
   unify_type(TyE, TyV, Pos, E),
   type_expr(I, Ctx, Pos, TyI),
   unify_type(TyI, Ty, Pos, I).
type_expr(forall(X, List, E), Ctx, Pos, Ty) :-
   !,
   type_expr(fold_left(and, List, X, true, E), Ctx, Pos, Ty).
type_expr(exists(X, List, E), Ctx, Pos, Ty) :-
   !,
   type_expr(fold_left(or, List, X, false, E), Ctx, Pos, Ty).
type_expr(foldl(X, List, Op, Id, E), Ctx, Pos, Ty) :-
   !,
   type_expr(fold_left(Op, List, X, Id, E), Ctx, Pos, Ty).
type_expr(foldr(X, List, Op, Id, E), Ctx, Pos, Ty) :-
   !,
   type_expr(fold_right(Op, List, X, Id, E), Ctx, Pos, Ty).
type_expr(let(X, E, F), Ctx, Pos, Ty) :-
   !,
   (
      X = var(NameX, _)
   ->
      true
   ;
      throw(type_error('Should be a variable'(X)), Pos, let(X, E, F))
   ),
   type_expr(E, Ctx, Pos, TyE),
   type_expr(F, [NameX = TyE | Ctx], Pos, Ty).
type_expr(C, Ctx, Pos, Ty) :-
   callable(C),
   !,
   C =.. [F | Args],
   get_type(F, Pos, C, TyArgs, Ty),
   type_list(Args, Ctx, F, Pos, C, TyArgs).

type_list(L, Ctx, F, Pos, Expr, TyL) :-
   length(L, N),
   length(TyL, N0),
   (
      N = N0
   ->
      type_list_aux(L, Ctx, Pos, TyL)
   ;
      throw(type_error('Arity mismatch'(N, F, N0), Pos, Expr))
   ).

new_var_name(Atom) :-
   (
      r2cp_var_counter(C)
   ->
      retract(r2cp_var_counter(_))
   ;
      C = 0
   ),
   N is C + 1,
   assert(r2cp_var_counter(N)),
   number_codes(N, NCodes),
   append("_var_", NCodes, Codes),
   atom_codes(Atom, Codes).

type_list_aux([], _Ctx, _Pos, []).
type_list_aux([H | T], Ctx, Pos, [TyH0 | TyT]) :-
   type_expr(H, Ctx, Pos, TyH),
   unify_type(TyH, TyH0, Pos, H),
   type_list_aux(T, Ctx, Pos, TyT).

type_record(L = V, Ctx, Pos, Ty) :-
   type_expr(V, Ctx, Pos, TyV),
   Ty = [L = TyV].
type_record((R0, R1), Ctx, Pos, Ty) :-
   type_record(R0, Ctx, Pos, TyR0),
   type_record(R1, Ctx, Pos, TyR1),
   (
      member(L = _, TyR0),
      (
         member(L = _, TyR1)
      ->
         throw(type_error('Field defined twice'(L), Pos, (R0, R1)))
      ;
         fail
      )
   ;
      true
   ),
   append(TyR0, TyR1, Ty).

unify_type(Ty0, Ty1, Pos, Expr) :-
   (
      (
         var(Ty0)
      ;
         var(Ty1)
      )
   ->
      Ty0 = Ty1
   ;
      Ty0 = {R0},
      Ty1 = {R1}
   ->
      unify_record(R0, R1, Ty0, Ty1, Pos, Expr)
   ;
      Ty0 = [S0],
      Ty1 = [S1]
   ->
      unify_type(S0, S1, Pos, Expr)
   ;
      Ty0 = Ty1
   ->
      true
   ;
      throw(type_error('Incompatible types'(Ty0, Ty1), Pos, Expr))
   ).

unify_record(R0, R1, _Ty0, _Ty1, _Pos, _Expr) :-
   var(R0),
   !,
   R0 = R1.
unify_record([], R1, Ty0, Ty1, Pos, Expr) :-
   (
      R1 = []
   ->
      true
   ;
      R1 = [L = _ | _]
   ->
      throw(type_error('Missing field'(L, Ty0, Ty1), Pos, Expr))
   ).
unify_record([L = TyL0 | T0], R1, Ty0, Ty1, Pos, Expr) :-
   (
      select(L = TyL1, R1, T1)
   ->
      unify_type(TyL0, TyL1, Pos, Expr),
      unify_record(T0, T1, Ty0, Ty1, Pos, Expr)
   ;
      throw(type_error('Missing field'(L, Ty1, Ty0), Pos, Expr))
   ).
