package rules2cp.interpreter.java;

import choco.kernel.solver.ContradictionException;
import choco.kernel.solver.branch.AbstractBinIntBranching;
import choco.kernel.solver.branch.VarSelector;
import choco.kernel.solver.search.integer.ValSelector;
import choco.kernel.solver.search.integer.VarValPairSelector;
import choco.kernel.solver.variables.integer.IntDomainVar;

import java.util.logging.Level;

/**
 * A dichotomic branching example
 */
public class DichotomicBranching extends AbstractBinIntBranching {

    protected VarSelector varSelector;
    
    public DichotomicBranching(VarSelector varSel) {
	super();
        this.varSelector = varSel;
    }

    public String getDecisionLogMsg(int branchIndex) {
	switch(branchIndex) {
	case 1:
	    return "=<";
	case 2:
	    return ">";
	default:
	    return "??";
	}
    }
    
    /**
     * delegates to the var selector the choice of the  branching variable
     */
    public Object selectBranchingObject() throws ContradictionException {
        return varSelector.selectVar();
    }
    
    public void goDownBranch(Object x, int i) throws ContradictionException {
	super.goDownBranch(x, i);
        IntDomainVar var = (IntDomainVar) x;
	
        int bound = (var.getSup() + var.getInf()) / 2;
	
        switch (i) {
	case 1: {
	    var.getSolver().post(var.getSolver().leq(var, bound));
	    var.getSolver().propagate();	    
	    break;
	}
	case 2: {
	    var.getSolver().post(var.getSolver().gt(var, bound));
	    var.getSolver().propagate();	    
	    break;
	}
        }
    }

    public void goUpBranch(Object x, int i) throws ContradictionException {
	super.goUpBranch(x, i);
    }

}
