:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- ensure_loaded(geosr).

/*
a first translation of psa.rcp to geosr by hand
- no weight balancing (no reification yet)
- non polymorphic objects
*/

psa_geosr([O1_,O1_2,O1_3, O2_,O2_2,O2_3, O3_,O3_2,O3_3,
	   O4_,O4_2,O4_3, O5_,O5_2,O5_3, O6_,O6_2,O6_3,
	   O7_,O7_2,O7_3, O8_,O8_2,O8_3, O9_,O9_2,O9_3]) :-
	
	O1_ in 0..1203, O1_2 in 0..235, O1_3 in 0..239,
	O2_ in 0..1203, O2_2 in 0..235, O2_3 in 0..239,
	O3_ in 0..1203, O3_2 in 0..235, O3_3 in 0..239,
	O4_ in 0..1203, O4_2 in 0..235, O4_3 in 0..239,
	O5_ in 0..1203, O5_2 in 0..235, O5_3 in 0..239,
	O6_ in 0..1203, O6_2 in 0..235, O6_3 in 0..239,
	O7_ in 0..1203, O7_2 in 0..235, O7_3 in 0..239,
	O8_ in 0..1203, O8_2 in 0..235, O8_3 in 0..239,
	O9_ in 0..1203, O9_2 in 0..235, O9_3 in 0..239,
	
	geosr([object(1, 1, [0,0,0],[]),
	       object(2, 4, [O1_,O1_2,O1_3], [weight-413]),
	       object(3, 5, [O2_,O2_2,O2_3], [weight-463]),
	       object(4, 5, [O3_,O3_2,O3_3], [weight-842]),
	       object(5, 3, [O4_,O4_2,O4_3], [weight-422]),
	       object(6, 4, [O5_,O5_2,O5_3], [weight-266]),
	       object(7, 4, [O6_,O6_2,O6_3], [weight-321]),
	       object(8, 2, [O7_,O7_2,O7_3], [weight-670]),
	       object(9, 6, [O8_,O8_2,O8_3], [weight-440]),
	       object(10,7, [O9_,O9_2,O9_3], [weight-325])],
	      
	      [sbox(1, [0,0,0], [1203, 235, 239]),
	       sbox(2, [0,0,0], [224, 224, 222]),
	       sbox(3, [0,0,0], [224, 224, 148]),
	       sbox(4, [0,0,0], [224, 224, 111]),
	       sbox(5, [0,0,0], [224, 224, 74]),
	       sbox(6, [0,0,0], [155, 224, 222]),
	       sbox(7, [0,0,0], [112, 224, 148])],
	      
	      [(origin(O,D) ---> O^x(D)),
	       
	       (size(S,D) ---> S^l(D)),
	       
	       (end(O,D) ---> origin(O, D) + fold(S, sboxes([O^sid]), +, 0, size(S, D))),
	       
	       (overlap_sym(O1, O2, D) --->
		   end(O1,D) #> origin(O2,D) #/\
	           end(O2,D) #> origin(O1,D)),
	       
	       (contains(O1,O2,D) --->
		   origin(O1,D) #=< origin(O2,D) #/\ 
	           end(O2,D) #=< end(O1,D)),
	       
	       (non_overlapping(Items, Dims) --->
		   forall(O1, Items, 
		      forall(O2, Items,
			     O1^oid #< O2^oid #=>
			    #\ overlap(O1,O2,Dims)))),
	       
	       (overlap(O1, O2, Dims) ---> forall(D, Dims, overlap_sym(O1, O2, D))),
	       
	       (containmentAE(Items, Bins, Dims) --->
		   forall(I, Items, 
		      exists(B, Bins, 
		         forall(D, Dims, contains(B,I,D))))),
	       
	       (bin_packing(Items, Bins, Dims) --->
		   containmentAE(Items, Bins, Dims) #/\
	           non_overlapping(Items, Dims)),
	       
	       (gravity(Items) ---> 
		   forall(O1, Items,
			 origin(O1, 3) #= 0 #\/
			 exists(O2, Items, O1^oid #\= O2^oid #/\ on_top(O1, O2)))),
	       
	       (on_top(O1, O2) ---> overlap(O1, O2, [1,2]) #/\ met_by(O1, O2, 3)),
	       
	       (met_by(A, B, D) ---> end(B, D) #= origin(A, D)),
	       
	       (weight_stacking(Items) --->
		   forall(O1,  Items,
			  forall(O2, Items,			
				(O1^weight #> O2^weight #/\ origin(O1,3) #>= end(O2,3))
				#=>
				#\ overlap(O1,O2,[1,2])) )),
	       
	       (lighter(O1, O2) ---> O1^weight #=< O2^weight),
	       
	       (stack_oversize(Items, Length) --->
		   forall(O1, Items,
			  forall(O2, Items,			
				(overlap(O1, O2, [1,2]) #/\ O1^oid #\= O2^oid)
				#=> 
				forall(D, [1,2], oversize(O1, O2, D) #=< Length)))),
	       
	       (oversize(O1, O2, D) ---> 
	           max( max( origin(O1, D), origin(O2, D)) - min( origin(O1, D), origin(O2, D)),
			max( end(O1, D), end(O2, D)) - min( end(O1, D), end(O2, D)))),
	       
	       (psa_bin_packing(Items, Bins, Dims) --->		   		   
		   gravity(Items) #/\
	           weight_stacking(Items) #/\
	           stack_oversize(Items, 10) #/\
	           bin_packing(Items, Bins, Dims)),
	       
	       psa_bin_packing(objects([2,3,4,5,6,7,8,9,10]), objects([1]), [1,2,3])]),

	fd_statistics,
	labeling([], [O1_,O1_2,O1_3, O2_,O2_2,O2_3, O3_,O3_2,O3_3,
		      O4_,O4_2,O4_3, O5_,O5_2,O5_3, O6_,O6_2,O6_3,
		      O7_,O7_2,O7_3, O8_,O8_2,O8_3, O9_,O9_2,O9_3]);
	fd_statistics.

/*
container(1, tc40, 1203, 235, 239).
object(1, cV8269, 224, 224, 111, 413).
object(2, pF8370, 224, 224, 74, 463).
object(3, cV8371, 224, 224, 74, 842).
object(4, cV8142, 224, 224, 148, 422).
object(5, cV8243, 224, 224, 111, 266).
object(6, cV8258, 224, 224, 111, 321).
object(7, pF8035, 224, 224, 222, 670).
object(8, pF7037, 155, 224, 222, 440).
object(9, cF6120, 112, 224, 148, 325).

object(1) : <0,0,0>
object(2) : <224,0,0>
object(3) : <448,0,0>
object(4) : <224,0,74>
object(5) : <0,0,111>
object(6) : <448,0,74>
object(7) : <672,0,0>
object(8) : <896,0,0>
object(9) : <1051,0,0>
*/